/*
 * rshaper.c
 *
 * Copyright (c) 1998,1999  Alessandro Rubini <rubini@prosa.it>
 *
 * This simple module intercepts netif_rx to lower host bandwidth
 * Tested with two ne2000 devices, see the accompanying 8390.patch
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#undef PDEBUG             /* undef it, just in case */
#ifdef RSHAPER_DEBUG
#  ifdef __KERNEL__
     /* This one if debugging is on, and kernel space */
#    define PDEBUG(fmt, args...) printk( KERN_DEBUG "rshaper: " fmt, ## args)
#  else
     /* This one for user space */
#    define PDEBUG(fmt, args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#  define PDEBUG(fmt, args...) /* not debugging: nothing */
#endif

#undef PDEBUGG
#define PDEBUGG(fmt, args...) /* nothing: it's a placeholder */

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/tcp.h>         /* struct tcphdr */
#include <linux/skbuff.h>

#include "rshaper.h" 

extern int (*net_shaper_rx_hook)(struct sk_buff *skb); /* in the eth driver */

/* ------------------------------------------------ Data Structures */

/* Each queued packet */
struct rshaper_packet {
    struct sk_buff *skb;
    unsigned long timestamp;
    struct rshaper_packet *next;
};

/* Each queue */
struct rshaper_queue {
    struct rshaper_packet *head; /* first packet, to dequeue */
    struct rshaper_packet *tail; /* last packet, to enqueue */
    struct timer_list timer;   /* use this to dequeue */
    unsigned long nexttick;      /* foreseen next tick */
    int approx;                  /* leftover bytes */
};

/* Each host or host group */
struct rshaper_host {
    u32 ip;
    u32 mask;
    struct rshaper_queue *tx_queue;
    struct rshaper_queue *rx_queue;
    struct rshaper_host *next;
    int bytes_per_tick;
    int timeout; /* measured in jiffies */
};    

/* ------------------------------------------------ Queue functions */

int enqueue_packet(struct sk_buff *skb, struct rshaper_queue *q,
                              int bytes_per_tick)
{
    struct rshaper_packet *pkt;
    int size = skb->tail - skb->data;
    unsigned long flags;
    int timerflag=0;

    /* if the queue is empty, transmit it now */
    if (q->nexttick <= jiffies) {
        PDEBUG("enqueue: now\n");
        netif_rx(skb);
        if (q->nexttick < jiffies)
            q->approx = 0;
        size += q->approx;
        q->nexttick = jiffies + size / bytes_per_tick;
        q->approx = size % bytes_per_tick;
        return 0;
    }

    /* allocate a structure */
    pkt = kmalloc(sizeof(struct rshaper_packet), GFP_ATOMIC);
    if (!pkt)
        return -ENOMEM;

    /* fill it */
    pkt->skb = skb;
    pkt->next = NULL;
    pkt->timestamp = q->nexttick;
    save_flags(flags); cli();

    PDEBUG("enqueue: %04i+%03i bytes, %3i tick, %3i left (delay %li j)\n",
           size, q->approx, (size+q->approx)/bytes_per_tick,
           (size+q->approx)%bytes_per_tick,
           q->nexttick - jiffies);
    size += q->approx;
    q->nexttick += size / bytes_per_tick;
    q->approx = size % bytes_per_tick;

    if (q->head) {
        PDEBUGG("updating tail (h %p, t %p)\n",q->head,q->tail);
        q->tail->next = pkt;
        q->tail = pkt;
    } else {
        PDEBUGG("registering the timer to run at %li\n",pkt->timestamp);
        q->head = q->tail = pkt;
        q->timer.expires = pkt->timestamp;
        timerflag++;
    }
    restore_flags(flags);
    if (timerflag)
        add_timer(&q->timer);
    return 0;
}

/* This is called by the timer */
void dequeue_packet(unsigned long data)
{
    struct rshaper_queue *q = (struct rshaper_queue *)data;
    struct rshaper_packet *p;

    PDEBUGG("timer running: now %li\n",jiffies);

    /* consume expired packets */
    while ((p = q->head) && (p->timestamp <= jiffies)) {
        netif_rx(p->skb);
        q->head = p->next;
        if (q->tail == p)
            q->tail = NULL;
        kfree(p);
    }
    /* if some more is there, schedule it */
    if (p) {
        q->timer.expires = p->timestamp;
        add_timer(&q->timer);
    }
}


/* ------------------------------------------------ Host Management */

struct rshaper_host *first_host;

struct rshaper_host *rshaper_find_host (u32 ip, struct rshaper_host ***prevp)
{
    struct rshaper_host *host, **prev = &first_host;

    PDEBUGG("Finding host %08lx\n",(unsigned long)ip);
    while (*prev && ((*prev)->ip != ip))
        prev = &((*prev)->next);

    if (!*prev) {
        *prevp = NULL;
        return NULL;
    }
    host = *prev;
    *prevp = prev; 
    return host;
}

int rshaper_add_host(u32 ip, u32 mask, int bytes_per_tick, int timeout)
{
    struct rshaper_host *host;
    
    host = kmalloc(sizeof(struct rshaper_host), GFP_KERNEL);
    if (!host) return -ENOMEM;

    PDEBUG("Adding host %08lx, %i bpt, %i sec\n",
           (unsigned long)ip, bytes_per_tick, timeout);

    host->ip = ip;
    host->mask = mask;
    host->bytes_per_tick = bytes_per_tick;
    host->timeout = timeout;
    host->tx_queue = kmalloc(sizeof(struct rshaper_queue), GFP_KERNEL);
    host->rx_queue = kmalloc(sizeof(struct rshaper_queue), GFP_KERNEL);

    if (!host->tx_queue || !host->rx_queue) {
        if (host->tx_queue) kfree(host->tx_queue);
        if (host->rx_queue) kfree(host->rx_queue);
        kfree(host);
        return -ENOMEM;
    }
    memset(host->rx_queue, 0, sizeof(struct rshaper_queue));
    init_timer(&host->rx_queue->timer);
    host->rx_queue->timer.data = (unsigned long)host->rx_queue;
    host->rx_queue->timer.function = dequeue_packet;

    memset(host->tx_queue, 0, sizeof(struct rshaper_queue));
    init_timer(&host->tx_queue->timer);
    host->tx_queue->timer.data = (unsigned long)host->tx_queue;
    host->tx_queue->timer.function = dequeue_packet;

    host->next = first_host;
    first_host = host;
    PDEBUGG("Registered with tx %p, rx %p\n", host->tx_queue, host->rx_queue);
    return 0;
    
}

void rshaper_kfree_host(struct rshaper_host *host)
{
    struct rshaper_packet *packet, *nextp;

    del_timer(&host->tx_queue->timer);
    del_timer(&host->rx_queue->timer);

    for (packet = host->tx_queue->head; packet; ) {
        nextp = packet->next;
        PDEBUGG("free packet %p (next %p)\n", packet, nextp);
        kfree(packet);
        packet = nextp;
    }
    kfree(host->tx_queue);
    
    for (packet = host->rx_queue->head; packet; ) {
        nextp = packet->next;
        PDEBUGG("free packet %p (next %p)\n", packet, nextp);
        kfree(packet);
        packet = nextp;
    }
    kfree(host->rx_queue);
    PDEBUG("free host %p\n", host);
    kfree(host);
}


/* ------------------------------------------------ Main Hook */

int rshaper_rx_hook(struct sk_buff *skb)
{
    /* no need to skip eth header, as this is after eth_type_trans() */
    struct iphdr *ih = (struct iphdr *)(skb->data);
    struct rshaper_host *ptr;
    u32 saddr = ih->saddr;
    u32 daddr = ih->daddr;

    PDEBUGG("packet: from %08lx to %08lx\n",(long)ih->saddr, (long)ih->daddr);

    /*
     * Ok, now we must check out our list of managed hosts, and delay
     * reception if found
     */
    for (ptr = first_host; ptr; ptr = ptr->next) {
        if (ptr->ip == (saddr & ptr->mask)) {
            PDEBUGG("Enqueue tx: %08lx\n", (long)saddr);
            if (ptr->tx_queue->nexttick <= jiffies + ptr->timeout) {
                enqueue_packet(skb, ptr->tx_queue, ptr->bytes_per_tick);
            } else {
                PDEBUG("Tx queue too long (%li+%i < %li, dropped\n",
		       jiffies, ptr->timeout, ptr->tx_queue->nexttick);
                skb->free = 1;
                kfree_skb(skb, 0);
            }
            return 0;
        }
        if (ptr->ip == (daddr & ptr->mask))  {
            PDEBUGG("Enqueue rx: %08lx\n", (long)daddr);
            if (ptr->rx_queue->nexttick <= jiffies + ptr->timeout) {
                enqueue_packet(skb, ptr->rx_queue, ptr->bytes_per_tick);
            } else {
                PDEBUG("Rx queue too long (%li+%i < %li, dropped\n",
		       jiffies, ptr->timeout, ptr->rx_queue->nexttick);
                skb->free = 1;
                kfree_skb(skb, 0);
            }
            return 0;
        }
    }

    netif_rx(skb); /* no delay, receive it right now */
    return 0;
}

/* ------------------------------------------------ User Cfg */

/*
 * There are several ways to talk to user space, devices, network interfaces,
 * /proc/entry points... Here I chose to use ioctl() on a network interface.
 *
 * SIOCDEVPRIVATE on rshaper is used to set host information,
 * SIOCDEVPRIVATE+1 is used to retrieve information.
 */
int rshaper_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
{
    struct rshaper_userinfo info,
         *uptr = (struct rshaper_userinfo *)ifr->ifr_data;
    struct rshaper_host *host, **prev;
    int i;

    if (!suser()) return -EPERM;

    if (verify_area(VERIFY_READ, uptr, sizeof(*uptr)))
        return -EINVAL;
    memcpy_fromfs(&info, uptr, sizeof(info));
    info.timeout *= HZ; /* measure in jiffies */
    PDEBUG("ioctl: host %08lx, count %i, to %i jiffies\n",(long)info.ip,
           (int)info.bytes_per_second, info.timeout);

    switch(cmd) {
      case SIOCRSHAPERSETINFO: /* set or clear one host */
        host = rshaper_find_host((u32)(info.ip), &prev);
        if (host) {
            if (info.bytes_per_second)  {
                /* turn byterate into bytes/tick, use nearest value */
                host->bytes_per_tick = (info.bytes_per_second+HZ/2)/HZ;
                host->timeout = info.timeout;
            } else { /* remove host */
                *prev = host->next;
                rshaper_kfree_host(host);
            }
        }
        else if (info.bytes_per_second) {
            rshaper_add_host(info.ip, info.mask,
			     (info.bytes_per_second+HZ/2)/HZ,
                             info.timeout);
        }
        return 0;

      case SIOCRSHAPERGETINFO: /* return information about the host */
        if (verify_area(VERIFY_READ, uptr, sizeof(*uptr)))
            return -EINVAL;

        for (host = first_host, i=0; host && i < info.ip;
             host = host->next, i++)
            /* nothing */;
        if (!host) {return -ENODEV;} /* no more hosts */
        info.ip = (__u32)(host->ip);
        info.mask = (__u32)(host->mask);
        info.bytes_per_second = HZ * host->bytes_per_tick;
        info.timeout = (host->timeout + HZ/2)/HZ; /* jiffies to seconds */
        PDEBUG("ioctl-ret: host %08lx (%08lx), count %i, to %i\n",
               (long)info.ip, (long)info.mask,
	       info.bytes_per_second, info.timeout);
        memcpy_tofs(uptr, &info, sizeof(info));
        return 0;

      default:
        return -EINVAL;
    }
    return 0;
}

/* ------------------------------------------------ The Network Device */

int rshaper_init_dev(struct device *dev)
{
    /* nothing more than ioctl */
    dev->do_ioctl = rshaper_ioctl;
    return 0;
}


struct device rshaper_dev = {
    "rshaper",
    0, 0, 0, 0,  /* shmem addresses */
    0x000,       /* ioport */
    0,           /* irq line */
    0, 0, 0,     /* various flags: init to 0 */
    NULL,        /* next ptr */
    rshaper_init_dev, /* init function, fill other fields with NULL's */
};    


/* ------------------------------------------------ Module Stuff */

int init_module(void)
{
    int result = register_netdev(&rshaper_dev);

    if (result) return result;
    net_shaper_rx_hook = rshaper_rx_hook;
    return 0;
}

void cleanup_module(void)
{
    struct rshaper_host *host, *next;

    net_shaper_rx_hook = NULL; /* de-hook */

    for (host = first_host; host; ) { /* free it all */
        next = host->next;
        rshaper_kfree_host(host);
        host = next;
    }
    unregister_netdev(&rshaper_dev);
}


